#ifndef CONFIG_H
#define CONFIG_H

// Maximum buffersize for input lines (a line may contain MAXLINE-2 chars).
#define MAXLINE 65535

// No special options
#define OPT_NONE 0

// Records may not span lines
#define OPT_NOFOLD 1

// Strip leading and trailing spaces/tabs
#define OPT_STRIP 2

// Ignore first line of input
#define OPT_IGHEADER 4

// Delete newlines from fields before expanding variables
#define OPT_NOBREAK 8 

/**
 * Config structure
 */
struct conf {
  char delim;		// Delimeter character
  char* tmplfile;	// File, containing the template
  char** fields;	// List of field names
  char* miss;		// Value for missing fields
  char* empty;		// Value for empty fields
  int opts;		// Boolean options (see OPT_* macros)
} cfg;

/**
 * We are reading from here
 */
FILE* in;

/**
 * And writing to there
 */
FILE* out;

/**
 * Fill the config structure with default values
 */
void setDefaults();

/**
 * Insert a value entry pair into the config structure
 * @param key config directive
 * @param val config argument (will be kept).
 */   
void putEntry(const char* key, char* val);

/**
 * Read config file (sorta main() of this module)
 * @param fname filename of the config file
 */
int readConf(const char* fname);

#endif
