/*
    Copyright (C) 2003, Patrick Ahlbrecht

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <getopt.h>
#include <string.h>

#include "util.h"
#include "config.h"


// Set program version               
char* version=VERSION;

/**
 * Contents of the template file
 */
char *tmpl;

/**
 * Strip trailing and leading characters
 * @param ptr pointer to the string
 * @param ch char to strip from beginning and end of ptr
 */
void strip(char* ptr,const char ch) {
  int x=strlen(ptr);
  int y=0;
  
  if(x==0) return;
  x--;
  
  while(ptr[x]==ch && x>=0) { 
    ptr[x]='\0';
    x--;
  }
  
  x=0;
  
  while(ptr[x]==ch) { y++; x++;}
  memmove(ptr,ptr+y,strlen(ptr));
}


/**
 * Expand input line
 * @param line one line of input
 * @return NULL terminated list of field values
 */
char** parseLine(const char* line) {
  int x=0;
  int y=0;
  int z=0;
  int opn=FALSE;
  char** split;
  
  if(line==NULL || line[0]=='\0') {
    split=(char**)calloc(1,sizeof(char*));
    oomtest(split);
    return split;
  }
  
  for(x=0;x<=strlen(line);x++) {if (line[x]==cfg.delim) y++;}
  split=(char**)calloc(y+2,sizeof(char*));
  oomtest(split);
  
  y=0;
  
  for(x=0;x<=strlen(line);x++) {
    if(x<=strlen(line)-2 &&line[x]=='"' && line[x+1]=='"' && line[x+2]=='"') {
      x=x+2;
    }
    else if(line[x]=='"') opn=!opn;
    
    if(!opn && (line[x]==cfg.delim || x==strlen(line)-1) ) {
      split[z]=(char*)calloc(x-y+2,sizeof(char));
      oomtest(split[z]);
      if (x<strlen(line)-1) memcpy(split[z],line+y,x-y);
      else memcpy(split[z],line+y,x-y+1);
      z++;
      y=x+1;
    }
  }
  
  if(split[z-1][strlen(split[z-1])-1]==cfg.delim) 
    split[z-1][strlen(split[z-1])-1]='\0';
  
  x=0;
  while(split[x]!=NULL) {
    if ((cfg.opts&OPT_STRIP) == OPT_STRIP) {
      strip(split[x],' ');
      strip(split[x],'\t');
    }
    if ((cfg.opts&OPT_NOBREAK) == OPT_NOBREAK) {
      split[x]=expand(split[x],"\n","");
    }
    strip(split[x],'"');
    split[x]=expand(split[x],"\"\"\"","\"");
    x++;
  }
  return split;
}

/**
 * Determnine, whether a line is "open", that is contains linebreaks as value
 * @param line line to check
 * @return TRUE traling newline is part of the value or line=="";
 */
int isOpen(const char* line) {
  int x=0;
  int opn=FALSE;
  
  for(x=0;x<=strlen(line);x++) {
    if(x<=strlen(line)-2 &&line[x]=='"' && line[x+1]=='"' && line[x+2]=='"') {
      x=x+2;
    }
    else if(line[x]=='"') opn=!opn;
  }
  return opn;
}

/**
 * Create output
 * @param vals the fields of the CSV file.
 */
void writeOutput(const char** vals) {
  int x=0;
  char* tmp;
  
  tmp=cpyStr(tmpl);
  
  while(cfg.fields[x]!=NULL) {
    if(vals[x]==NULL) break;
    if(vals[x][0]=='\0') tmp=expand(tmp,cfg.fields[x],cfg.empty);
    else tmp=expand(tmp,cfg.fields[x],vals[x]);
    x++;
  }
  
  while(cfg.fields[x]!=NULL) {
    tmp=expand(tmp,cfg.fields[x],cfg.miss);
    x++;
  }
  
  fwrite(tmp,strlen(tmp),sizeof(char),out);
  free(tmp);
}

/**
 * Read lines
 */
void readLines() {
  char ibuf[MAXLINE]; // "inputbuffer"
  char *bbuf=NULL;    // "backbuffer"
  int first=TRUE;
  
  bbuf=NULL;
  
  while (fgets(ibuf,(int)sizeof(ibuf)-2*sizeof(char),in) ) {
    if((cfg.opts&OPT_NOFOLD)==OPT_NOFOLD && isOpen(ibuf)) {
      ibuf[strlen(ibuf)+1]='\0';
      if(ibuf[strlen(ibuf)-1]=='\n') ibuf[strlen(ibuf)-1]='"';
      else ibuf[strlen(ibuf)]='"';
    }
    if (bbuf==NULL || isOpen(bbuf)) {
      int len=0;
      if(bbuf!=NULL) len=strlen(bbuf);
      bbuf=(char*)realloc(bbuf,(strlen(ibuf)+len+1)*sizeof(char));
      oomtest(bbuf);
      memcpy(bbuf+len,ibuf,strlen(ibuf)+sizeof(char));
    }
    if(!isOpen(bbuf)) {
      int i=0;
      char** fields;
      
      if(bbuf[strlen(bbuf)-1]=='\n') bbuf[strlen(bbuf)-1]='\0';
      fields=parseLine(bbuf);
      free(bbuf);
      if( ! ((cfg.opts&OPT_IGHEADER)==OPT_IGHEADER && first==TRUE) ) {
        writeOutput((const char**)fields);
      }
      freeList((void**)fields);
      bbuf=NULL;
      first=FALSE;
    }
  }    
}

/**
 * Print Usage and exit
 * @param name argv[0] of main
 */
void printUsage(const char* name) {
  
  printf("%s [OPTIONS] ...\n\
  \nConvert files from CSV format using a template\n\n\
  -h\t\t\tPrint usage\n\
  -c <file>\t\tUse <file> as control file (required)\n\
  -i <file>\t\tRead CSV data from <file>, instead of stdin\n\
  -o <file>\t\tWrite output to <file>, instead of stdout\n\
  \n",name);
  exit(EXIT_SUCCESS);
}

int main(int argc,char**argv) {
  int ch;
  int x=0;
  int y=0;
  char* cfgfile=NULL;
  char* bname=(char*)basename(argv[0]);
  
  in=stdin;
  out=stdout;
  
  while ((ch = getopt(argc, argv, "hi:o:c:")) !=-1) {
    switch (ch) {
      case 'h' : {
        printUsage(bname);
        break;
      }
      
      case 'c': {
        cfgfile=optarg;
        break;
      }
      
      case 'o': {
        out=fopen(optarg,"w");
        if (out==NULL) {
          perror(optarg);
          return EXIT_FAILURE;
        }
        break;
      }
      
      case 'i': {
        in=fopen(optarg,"r");
        if (in==NULL) {
          perror(optarg);
          return EXIT_FAILURE;
        }
        break;
      }
    }
  }
  
  if(cfgfile==NULL) printUsage(bname);
  
  readConf(cfgfile);
  
  if(cfg.tmplfile==NULL) {
    fprintf(stderr,"Error: No template file specified\n");
    return EXIT_FAILURE;
  }
  
  if ((tmpl=readFile(cfg.tmplfile))==NULL) {
    perror(cfg.tmplfile);
    return EXIT_FAILURE;
  }
  
  readLines(in,out);
  return EXIT_SUCCESS;
}
